@@ -22,20 +22,37 @@
 #
 # SPDX-License-Identifier: LGPL-2.0-or-later
 
-import sys
+from __future__ import annotations
+
 import inspect
 import itertools
 import weakref
-from ..overrides import override
-from ..module import get_introspection_module
+import typing
 
-from gi.repository import GLib
+from gi.overrides import override
+from gi.repository import GLib, GObject
 
-
-Gst = get_introspection_module('Gst')
+# Typing relies on https://github.com/pygobject/pygobject-stubs.
+if typing.TYPE_CHECKING:
+    # Import stubs for type checking this file.
+    #
+    # This causes some weirdness because stubs contains overridden APIs
+    # signatures. For example when using Gst.Bin.add() here, we mean to call the
+    # g-i generated API which does not have the same signature as our override
+    # Bin.add(). The type checker will use signature from stubs which is our
+    # override signature.
+    #
+    # For similar reason, make sure the type of arguments and return values
+    # are e.g. `Element` and not `Gst.Element`. Once copied into stubs,
+    # `Gst.Element` would not be defined, Gst module cannot refer to itself.
+    from gi.repository import Gst
+else:
+    from gi.module import get_introspection_module
+    Gst = get_introspection_module('Gst')
 
 __all__ = []
 
+
 if Gst._version == '0.10':
     import warnings
     warn_msg = "You have imported the Gst 0.10 module.  Because Gst 0.10 \
@@ -53,12 +70,14 @@ python module to use with Gst 0.10"
 class URIHandler(Gst.URIHandler):
     pass
 
-URIHandler = override(URIHandler)
+
+override(URIHandler)
 __all__.append('URIHandler')
 
+
 class Element(Gst.Element):
     @staticmethod
-    def link_many(*args):
+    def link_many(*args: Element) -> None:  # type: ignore[override]
         '''
         @raises: Gst.LinkError
         '''
@@ -67,35 +86,135 @@ class Element(Gst.Element):
                 raise LinkError(
                     'Failed to link {} and {}'.format(pair[0], pair[1]))
 
-Element = override(Element)
+
+override(Element)
 __all__.append('Element')
 
 
 class Bin(Gst.Bin):
-    def __init__(self, name=None):
+    def __init__(self, name: typing.Optional[str] = None):
         Gst.Bin.__init__(self, name=name)
 
-    def add(self, *args):
+    def add(self, *args: Element) -> None:  # type: ignore[override]
         for arg in args:
-            if not Gst.Bin.add(self, arg):
+            if not Gst.Bin.add(self, arg):  # type: ignore[func-returns-value]
                 raise AddError(arg)
 
-    def make_and_add(self, factory_name, instance_name=None):
+    def make_and_add(self, factoryname: str, name: typing.Optional[str] = None) -> Element:
         '''
         @raises: Gst.AddError
         '''
-        elem = Gst.ElementFactory.make(factory_name, instance_name)
+        elem = Gst.ElementFactory.make(factoryname, name)
         if not elem:
             raise AddError(
-                'No such element: {}'.format(factory_name))
+                'No such element: {}'.format(factoryname))
+        assert isinstance(elem, Element)  # Tell mypy we actually have our override subclass
         self.add(elem)
         return elem
 
 
-Bin = override(Bin)
+override(Bin)
 __all__.append('Bin')
 
-class Caps(Gst.Caps):
+
+class NotWritableMiniObject(Exception):
+    pass
+
+
+__all__.append('NotWritableMiniObject')
+
+
+class MiniObject:
+    def make_writable(self) -> MiniObject:
+        return _gi_gst.mini_object_make_writable(self)
+
+    def is_writable(self) -> bool:
+        return _gi_gst.mini_object_is_writable(self)
+
+    @property
+    def flags(self) -> Gst.MiniObjectFlags:
+        return _gi_gst.mini_object_flags(self)
+
+    @flags.setter
+    def flags(self, flags: Gst.MiniObjectFlags) -> None:
+        _gi_gst.mini_object_set_flags(self, flags)
+
+
+class NotWritableQuery(Exception):
+    pass
+
+
+__all__.append('NotWritableQuery')
+
+
+class Query(MiniObject, Gst.Query):
+    def get_structure(self) -> typing.Optional[Structure]:
+        s = _gi_gst.query_get_structure(self)
+        return s._set_parent(self) if s is not None else None
+
+    def writable_structure(self) -> StructureWrapper:
+        return StructureWrapper(_gi_gst.query_writable_structure(self)._set_parent(self))
+
+
+override(Query)
+__all__.append('Query')
+
+
+class NotWritableEvent(Exception):
+    pass
+
+
+__all__.append('NotWritableEvent')
+
+
+class Event(MiniObject, Gst.Event):
+    def get_structure(self) -> typing.Optional[Structure]:
+        s = _gi_gst.event_get_structure(self)
+        return s._set_parent(self) if s is not None else None
+
+    def writable_structure(self) -> StructureWrapper:
+        return StructureWrapper(_gi_gst.event_writable_structure(self)._set_parent(self))
+
+
+override(Event)
+__all__.append('Event')
+
+
+class NotWritableContext(Exception):
+    pass
+
+
+__all__.append('NotWritableContext')
+
+
+class Context(MiniObject, Gst.Context):
+    def get_structure(self) -> typing.Optional[Structure]:
+        s = _gi_gst.context_get_structure(self)
+        return s._set_parent(self) if s is not None else None
+
+    def writable_structure(self) -> StructureWrapper:
+        return StructureWrapper(_gi_gst.context_writable_structure(self)._set_parent(self))
+
+
+override(Context)
+__all__.append('Context')
+
+
+class NotWritableCaps(Exception):
+    pass
+
+
+__all__.append('NotWritableCaps')
+
+
+class NotWritableStructure(Exception):
+    pass
+
+
+__all__.append('NotWritableStructure')
+
+
+class Caps(MiniObject, Gst.Caps):
 
     def __nonzero__(self):
         return not self.is_empty()
@@ -103,9 +222,11 @@ class Caps(Gst.Caps):
     def __new__(cls, *args):
         if not args:
             return Caps.new_empty()
-        elif len(args) > 1:
+        if len(args) > 1:
             raise TypeError("wrong arguments when creating GstCaps object")
-        elif isinstance(args[0], str):
+
+        assert len(args) == 1
+        if isinstance(args[0], str):
             return Caps.from_string(args[0])
         elif isinstance(args[0], Caps):
             return args[0].copy()
@@ -130,16 +251,26 @@ class Caps(Gst.Caps):
     def __getitem__(self, index):
         if index >= self.get_size():
             raise IndexError('structure index out of range')
-        return self.get_structure(index)
+
+        return Gst.Caps.get_structure(self, index)
 
     def __len__(self):
         return self.get_size()
 
-Caps = override(Caps)
+    def get_structure(self, index: int) -> typing.Optional[Structure]:
+        s = _gi_gst.caps_get_structure(self, index)
+        return s._set_parent(self) if s is not None else None
+
+    def writable_structure(self, index: int) -> StructureWrapper:
+        return StructureWrapper(_gi_gst.caps_writable_structure(self, index)._set_parent(self))
+
+
+override(Caps)
 __all__.append('Caps')
 
+
 class PadFunc:
-    def __init__(self, func):
+    def __init__(self, func: typing.Callable[..., Gst.FlowReturn]):
         self.func = func
 
     def __call__(self, pad, parent, obj):
@@ -159,23 +290,24 @@ class PadFunc:
 
         return res
 
+
 class Pad(Gst.Pad):
     def __init__(self, *args, **kwargs):
         super(Gst.Pad, self).__init__(*args, **kwargs)
 
-    def set_chain_function(self, func):
+    def set_chain_function(self, func: typing.Callable[..., Gst.FlowReturn]) -> None:
         self.set_chain_function_full(PadFunc(func), None)
 
-    def set_event_function(self, func):
+    def set_event_function(self, func: typing.Callable[..., Gst.FlowReturn]) -> None:
         self.set_event_function_full(PadFunc(func), None)
 
-    def set_query_function(self, func):
+    def set_query_function(self, func: typing.Callable[..., Gst.FlowReturn]) -> None:
         self.set_query_function_full(PadFunc(func), None)
 
     def query_caps(self, filter=None):
         return Gst.Pad.query_caps(self, filter)
 
-    def set_caps(self, caps):
+    def set_caps(self, caps: Gst.Caps) -> bool:
         if not isinstance(caps, Gst.Caps):
             raise TypeError("%s is not a Gst.Caps." % (type(caps)))
 
@@ -191,17 +323,19 @@ class Pad(Gst.Pad):
 
         return res
 
-    def link(self, pad):
+    def link(self, pad: Gst.Pad) -> Gst.PadLinkReturn:
         ret = Gst.Pad.link(self, pad)
         if ret != Gst.PadLinkReturn.OK:
             raise LinkError(ret)
         return ret
 
-Pad = override(Pad)
+
+override(Pad)
 __all__.append('Pad')
 
+
 class GhostPad(Gst.GhostPad):
-    def __init__(self, name, target=None, direction=None):
+    def __init__(self, name: str, target: typing.Optional[Gst.Pad] = None, direction: typing.Optional[Gst.PadDirection] = None):
         if direction is None:
             if target is None:
                 raise TypeError('you must pass at least one of target '
@@ -213,26 +347,39 @@ class GhostPad(Gst.GhostPad):
         if target is not None:
             self.set_target(target)
 
-    def query_caps(self, filter=None):
+    def query_caps(self, filter: typing.Optional[Gst.Caps] = None) -> Gst.Caps:
         return Gst.GhostPad.query_caps(self, filter)
 
-GhostPad = override(GhostPad)
+
+override(GhostPad)
 __all__.append('GhostPad')
 
+
 class IteratorError(Exception):
     pass
+
+
 __all__.append('IteratorError')
 
+
 class AddError(Exception):
     pass
+
+
 __all__.append('AddError')
 
+
 class LinkError(Exception):
     pass
+
+
 __all__.append('LinkError')
 
+
 class MapError(Exception):
     pass
+
+
 __all__.append('MapError')
 
 
@@ -248,63 +395,76 @@ class Iterator(Gst.Iterator):
 
             yield value
 
-Iterator = override(Iterator)
+
+override(Iterator)
 __all__.append('Iterator')
 
 
 class ElementFactory(Gst.ElementFactory):
 
     # ElementFactory
-    def get_longname(self):
+    def get_longname(self) -> typing.Optional[str]:
         return self.get_metadata("long-name")
 
-    def get_description(self):
+    def get_description(self) -> typing.Optional[str]:
         return self.get_metadata("description")
 
-    def get_klass(self):
+    def get_klass(self) -> typing.Optional[str]:
         return self.get_metadata("klass")
 
-    @classmethod
-    def make(cls, factory_name, instance_name=None):
-        return Gst.ElementFactory.make(factory_name, instance_name)
+    @staticmethod
+    def make(factoryname: str, name: typing.Optional[str] = None) -> typing.Optional[Element]:  # type: ignore[override]
+        elem = Gst.ElementFactory.make(factoryname, name)
+        assert elem is None or isinstance(elem, Element)  # Tell mypy we actually have our override subclass
+        return elem
+
 
 class Pipeline(Gst.Pipeline):
-    def __init__(self, name=None):
+    def __init__(self, name: typing.Optional[str] = None):
         Gst.Pipeline.__init__(self, name=name)
 
-Pipeline = override(Pipeline)
+
+override(Pipeline)
 __all__.append('Pipeline')
 
+
 class Structure(Gst.Structure):
     def __new__(cls, *args, **kwargs):
         if not args:
             if kwargs:
                 raise TypeError("wrong arguments when creating GstStructure, first argument"
                                 " must be the structure name.")
-            return Structure.new_empty()
+            struct = Structure.new_empty()
+            return struct
         elif len(args) > 1:
             raise TypeError("wrong arguments when creating GstStructure object")
         elif isinstance(args[0], str):
             if not kwargs:
-                return Structure.from_string(args[0])[0]
+                struct = Structure.from_string(args[0])[0]
+                return struct
             struct = Structure.new_empty(args[0])
             for k, v in kwargs.items():
                 struct[k] = v
 
             return struct
         elif isinstance(args[0], Structure):
-            return args[0].copy()
+            struct = args[0].copy()
+            return struct
 
         raise TypeError("wrong arguments when creating GstStructure object")
 
     def __init__(self, *args, **kwargs):
         pass
 
+    def __ptr__(self):
+        return _gi_gst._get_object_ptr(self)
+
     def __getitem__(self, key):
         return self.get_value(key)
 
-    def keys(self):
+    def keys(self) -> set[str]:
         keys = set()
+
         def foreach(fid, value, unused1, udata):
             keys.add(GLib.quark_to_string(fid))
             return True
@@ -315,17 +475,37 @@ class Structure(Gst.Structure):
     def __setitem__(self, key, value):
         return self.set_value(key, value)
 
-    def __str__(self):
+    def set_value(self, key: str, value: typing.Any) -> bool:
+        if not _gi_gst.structure_is_writable(self):
+            raise NotWritableStructure("Trying to write to a not writable structure."
+                                       " Make sure to use the right APIs to have access to structure"
+                                       " in a writable way.")
+
+        return Gst.Structure.set_value(self, key, value)
+
+    def __str__(self) -> str:
         return self.to_string()
 
-Structure = override(Structure)
+    def _set_parent(self, parent):
+        self.__parent__ = parent
+        return self
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, _type, _value, _tb):
+        self._set_parent(None)
+
+
+override(Structure)
 __all__.append('Structure')
 
-ElementFactory = override(ElementFactory)
+override(ElementFactory)
 __all__.append('ElementFactory')
 
+
 class Fraction(Gst.Fraction):
-    def __init__(self, num, denom=1):
+    def __init__(self, num: int, denom: int = 1):
         def __gcd(a, b):
             while b != 0:
                 tmp = a
@@ -377,7 +557,7 @@ class Fraction(Gst.Fraction):
         elif isinstance(other, int):
             return Fraction(self.num * other, self.denom)
         raise TypeError("%s is not supported, use Gst.Fraction or int." %
-                (type(other)))
+                        (type(other)))
 
     __rmul__ = __mul__
 
@@ -388,7 +568,7 @@ class Fraction(Gst.Fraction):
         elif isinstance(other, int):
             return Fraction(self.num, self.denom * other)
         return TypeError("%s is not supported, use Gst.Fraction or int." %
-                (type(other)))
+                         (type(other)))
 
     __div__ = __truediv__
 
@@ -405,7 +585,8 @@ class Fraction(Gst.Fraction):
     def __str__(self):
         return '%d/%d' % (self.num, self.denom)
 
-Fraction = override(Fraction)
+
+override(Fraction)
 __all__.append('Fraction')
 
 
@@ -427,14 +608,14 @@ class IntRange(Gst.IntRange):
 
     def __repr__(self):
         return '<Gst.IntRange [%d,%d,%d]>' % (self.range.start,
-                self.range.stop, self.range.step)
+                                              self.range.stop, self.range.step)
 
     def __str__(self):
         if self.range.step == 1:
             return '[%d,%d]' % (self.range.start, self.range.stop)
         else:
             return '[%d,%d,%d]' % (self.range.start, self.range.stop,
-                    self.range.step)
+                                   self.range.step)
 
     def __eq__(self, other):
         if isinstance(other, range):
@@ -443,9 +624,9 @@ class IntRange(Gst.IntRange):
             return self.range == other.range
         return False
 
-if sys.version_info >= (3, 0):
-    IntRange = override(IntRange)
-    __all__.append('IntRange')
+
+override(IntRange)
+__all__.append('IntRange')
 
 
 class Int64Range(Gst.Int64Range):
@@ -466,14 +647,14 @@ class Int64Range(Gst.Int64Range):
 
     def __repr__(self):
         return '<Gst.Int64Range [%d,%d,%d]>' % (self.range.start,
-                self.range.stop, self.range.step)
+                                                self.range.stop, self.range.step)
 
     def __str__(self):
         if self.range.step == 1:
             return '(int64)[%d,%d]' % (self.range.start, self.range.stop)
         else:
             return '(int64)[%d,%d,%d]' % (self.range.start, self.range.stop,
-                    self.range.step)
+                                          self.range.step)
 
     def __eq__(self, other):
         if isinstance(other, range):
@@ -482,8 +663,9 @@ class Int64Range(Gst.Int64Range):
             return self.range == other.range
         return False
 
+
 class Bitmask(Gst.Bitmask):
-    def __init__(self, v):
+    def __init__(self, v: int) -> None:
         if not isinstance(v, int):
             raise TypeError("%s is not an int." % (type(v)))
 
@@ -496,17 +678,16 @@ class Bitmask(Gst.Bitmask):
         return self.v == other
 
 
-Bitmask = override(Bitmask)
+override(Bitmask)
 __all__.append('Bitmask')
 
 
-if sys.version_info >= (3, 0):
-    Int64Range = override(Int64Range)
-    __all__.append('Int64Range')
+override(Int64Range)
+__all__.append('Int64Range')
 
 
 class DoubleRange(Gst.DoubleRange):
-    def __init__(self, start, stop):
+    def __init__(self, start: int | float, stop: int | float):
         self.start = float(start)
         self.stop = float(stop)
 
@@ -520,16 +701,16 @@ class DoubleRange(Gst.DoubleRange):
         return '(double)[%s,%s]' % (str(self.range.start), str(self.range.stop))
 
 
-DoubleRange = override(DoubleRange)
+override(DoubleRange)
 __all__.append('DoubleRange')
 
 
 class FractionRange(Gst.FractionRange):
-    def __init__(self, start, stop):
-        if not isinstance(start, Gst.Fraction):
+    def __init__(self, start: Fraction, stop: Fraction):
+        if not isinstance(start, Fraction):
             raise TypeError("%s is not a Gst.Fraction." % (type(start)))
 
-        if not isinstance(stop, Gst.Fraction):
+        if not isinstance(stop, Fraction):
             raise TypeError("%s is not a Gst.Fraction." % (type(stop)))
 
         if (float(start) >= float(stop)):
@@ -540,12 +721,13 @@ class FractionRange(Gst.FractionRange):
 
     def __repr__(self):
         return '<Gst.FractionRange [%s,%s]>' % (str(self.start),
-                str(self.stop))
+                                                str(self.stop))
 
     def __str__(self):
         return '(fraction)[%s,%s]' % (str(self.start), str(self.stop))
 
-FractionRange = override(FractionRange)
+
+override(FractionRange)
 __all__.append('FractionRange')
 
 
@@ -563,12 +745,13 @@ class ValueArray(Gst.ValueArray):
         return len(self.array)
 
     def __str__(self):
-        return '<' + ','.join(map(str,self.array)) + '>'
+        return '<' + ','.join(map(str, self.array)) + '>'
 
     def __repr__(self):
         return '<Gst.ValueArray %s>' % (str(self))
 
-ValueArray = override(ValueArray)
+
+override(ValueArray)
 __all__.append('ValueArray')
 
 
@@ -586,22 +769,81 @@ class ValueList(Gst.ValueList):
         return len(self.array)
 
     def __str__(self):
-        return '{' + ','.join(map(str,self.array)) + '}'
+        return '{' + ','.join(map(str, self.array)) + '}'
 
     def __repr__(self):
         return '<Gst.ValueList %s>' % (str(self))
 
-ValueList = override(ValueList)
+
+override(ValueList)
 __all__.append('ValueList')
 
+
+class TagList(Gst.TagList):
+    def __init__(self):
+        Gst.TagList.__init__(self)
+
+    def __getitem__(self, index):
+        if index >= self.n_tags():
+            raise IndexError('taglist index out of range')
+
+        key = self.nth_tag_name(index)
+        (res, val) = Gst.TagList.copy_value(self, key)
+        if not res:
+            raise KeyError(f"tag {key} not found")
+        return val
+
+    def __setitem__(self, key, value):
+        self.add(Gst.TagMergeMode.REPLACE, key, value)
+
+    def keys(self):
+        keys = set()
+
+        def foreach(list, fid, value, udata):
+            keys.add(fid)
+            return True
+
+        self.foreach(foreach, None, None)
+        return keys
+
+    def enumerate(self):
+        return map(lambda k: (k, Gst.TagList.copy_value(self, k)[1]), self.keys())
+
+    def __len__(self):
+        return self.n_tags()
+
+    def __str__(self):
+        return self.to_string()
+
+    def __repr__(self):
+        return '<Gst.TagList %s>' % (str(self))
+
+
+override(TagList)
+__all__.append('TagList')
+
 # From https://docs.python.org/3/library/itertools.html
 
 
-def pairwise(iterable):
+def pairwise(iterable: typing.Iterable[Element]) -> typing.Iterator[tuple[Element, Element]]:
     a, b = itertools.tee(iterable)
     next(b, None)
     return zip(a, b)
 
+
+class StructureWrapper:
+    """A Gst.Structure wrapper to force usage of a context manager.
+    """
+    def __init__(self, structure: Structure):
+        self.__structure = structure
+
+    def __enter__(self) -> Structure:
+        return self.__structure
+
+    def __exit__(self, _type, _value, _tb):
+        self.__structure._set_parent(None)
+
+
 class MapInfo:
     def __init__(self):
         self.memory = None
@@ -628,48 +870,101 @@ class MapInfo:
         if not self.__parent__.unmap(self):
             raise MapError('MappingError', 'Unmapping was not successful')
 
+
 __all__.append("MapInfo")
 
-class Buffer(Gst.Buffer):
 
-    def map_range(self, idx, length, flags):
+class Buffer(MiniObject, Gst.Buffer):
+    @property  # type: ignore[override]
+    def flags(self) -> Gst.BufferFlags:
+        return _gi_gst.mini_object_flags(self)
+
+    @flags.setter
+    def flags(self, flags: Gst.BufferFlags) -> None:
+        _gi_gst.mini_object_set_flags(self, flags)
+
+    @property
+    def dts(self) -> int:
+        return _gi_gst.buffer_get_dts(self)
+
+    @dts.setter
+    def dts(self, dts: int) -> None:
+        _gi_gst.buffer_set_dts(self, dts)
+
+    @property
+    def pts(self) -> int:
+        return _gi_gst.buffer_get_pts(self)
+
+    @pts.setter
+    def pts(self, pts: int) -> None:
+        _gi_gst.buffer_set_pts(self, pts)
+
+    @property
+    def duration(self) -> int:
+        return _gi_gst.buffer_get_duration(self)
+
+    @duration.setter
+    def duration(self, duration: int) -> None:
+        _gi_gst.buffer_set_duration(self, duration)
+
+    @property
+    def offset(self) -> int:
+        return _gi_gst.buffer_get_offset(self)
+
+    @offset.setter
+    def offset(self, offset: int) -> None:
+        _gi_gst.buffer_set_offset(self, offset)
+
+    @property
+    def offset_end(self) -> int:
+        return _gi_gst.buffer_get_offset_end(self)
+
+    @offset_end.setter
+    def offset_end(self, offset_end: int) -> None:
+        _gi_gst.buffer_set_offset_end(self, offset_end)
+
+    def map_range(self, idx: int, length: int, flags: Gst.MapFlags) -> MapInfo:  # type: ignore[override]
         mapinfo = MapInfo()
         if (_gi_gst.buffer_override_map_range(self, mapinfo, idx, length, int(flags))):
             mapinfo.__parent__ = self
 
         return mapinfo
 
-    def map(self, flags):
+    def map(self, flags: Gst.MapFlags) -> MapInfo:  # type: ignore[override]
         mapinfo = MapInfo()
         if _gi_gst.buffer_override_map(self, mapinfo, int(flags)):
             mapinfo.__parent__ = self
 
         return mapinfo
 
-    def unmap(self, mapinfo):
+    def unmap(self, mapinfo: MapInfo) -> bool:  # type: ignore[override]
         mapinfo.__parent__ = None
         return _gi_gst.buffer_override_unmap(self, mapinfo)
 
-Buffer = override(Buffer)
+
+override(Buffer)
 __all__.append('Buffer')
 
+
 class Memory(Gst.Memory):
 
-    def map(self, flags):
+    def map(self, flags: Gst.MapFlags) -> MapInfo:  # type: ignore[override]
         mapinfo = MapInfo()
         if (_gi_gst.memory_override_map(self, mapinfo, int(flags))):
             mapinfo.__parent__ = self
 
         return mapinfo
 
-    def unmap(self, mapinfo):
+    def unmap(self, mapinfo: MapInfo) -> bool:  # type: ignore[override]
         mapinfo.__parent__ = None
         return _gi_gst.memory_override_unmap(self, mapinfo)
 
-Memory = override(Memory)
+
+override(Memory)
 __all__.append('Memory')
 
-def TIME_ARGS(time):
+
+def TIME_ARGS(time: int) -> str:
     if time == Gst.CLOCK_TIME_NONE:
         return "CLOCK_TIME_NONE"
 
@@ -677,9 +972,11 @@ def TIME_ARGS(time):
                                   (time / (Gst.SECOND * 60)) % 60,
                                   (time / Gst.SECOND) % 60,
                                   time % Gst.SECOND)
+
+
 __all__.append('TIME_ARGS')
 
-from gi.overrides import _gi_gst
+from gi.overrides import _gi_gst  # type: ignore[attr-defined]
 _gi_gst
 
 # maybe more python and less C some day if core turns a bit more introspection
@@ -694,10 +991,15 @@ Gst.fixme = _gi_gst.fixme
 Gst.memdump = _gi_gst.memdump
 
 # Make sure PyGst is not usable if GStreamer has not been initialized
+
+
 class NotInitialized(Exception):
     pass
+
+
 __all__.append('NotInitialized')
 
+
 def fake_method(*args):
     raise NotInitialized("Please call Gst.init(argv) before using GStreamer")
 
@@ -711,6 +1013,24 @@ for cname_klass in [o for o in inspect.getmembers(Gst) if isinstance(o[1], type(
                           for o in cname_klass[1].__dict__
                           if isinstance(cname_klass[1].__dict__[o], type(Gst.init))]))
 
+pre_init_functions = set([
+    "init",
+    "init_check",
+    "deinit",
+    "is_initialized",
+    "debug_add_log_function",
+    "debug_add_ring_buffer_logger",
+    "debug_remove_log_function",
+    "debug_remove_log_function_by_data",
+    "debug_remove_ring_buffer_logger",
+    "debug_set_active",
+    "debug_set_color_mode",
+    "debug_set_color_mode_from_string",
+    "debug_set_colored",
+    "debug_set_default_threshold",
+])
+
+
 def init_pygst():
     for fname, function in real_functions:
         if fname not in ["init", "init_check", "deinit"]:
@@ -723,37 +1043,47 @@ def init_pygst():
 
 def deinit_pygst():
     for fname, func in real_functions:
-        if fname not in ["init", "init_check", "deinit", "is_initialized"]:
+        if fname not in pre_init_functions:
             setattr(Gst, fname, fake_method)
     for cname_class, methods in class_methods:
         for mname, method in methods:
             setattr(cname_class[1], mname, fake_method)
 
+
 real_init = Gst.init
-def init(argv):
+
+
+def init(argv: typing.Optional[list[str]] = None) -> None:
     init_pygst()
 
     if Gst.is_initialized():
-        return True
+        return
 
     return real_init(argv)
 
+
 Gst.init = init
 
 real_init_check = Gst.init_check
-def init_check(argv):
+
+
+def init_check(argv: typing.Optional[list[str]] = None) -> typing.Tuple[bool, typing.Optional[list[str]]]:
     init_pygst()
     if Gst.is_initialized():
-        return True
+        return True, argv
 
     return real_init_check(argv)
 
+
 Gst.init_check = init_check
 
 real_deinit = Gst.deinit
-def deinit():
+
+
+def deinit() -> None:
     deinit_pygst()
-    return real_deinit()
+    real_deinit()
+
 
 def init_python():
     if not Gst.is_initialized():
@@ -761,6 +1091,7 @@ def init_python():
 
     init_pygst()
 
+
 Gst.deinit = deinit
 Gst.init_python = init_python
 
